-- Hook Chat Links
local Original_ChatEdit_InsertLink = ChatEdit_InsertLink
function ChatEdit_InsertLink(item, ...)
   local editbox = GFL_MainFrame_EditBox
   local focus = editbox:HasFocus()
   local visible = editbox:IsVisible()

   if item and focus and visible and IsShiftKeyDown() then
      editbox:Insert(item)
      return true
   end
   return Original_ChatEdit_InsertLink(item, ...)
end

-- Private Data
local ADDON_PREFIX = "GFL_Prefix"
local ADDON_NAME = "GuildFarmingList"

local Utils = GFL_UtilFuncs

local SecsPerMin = 60
local SecsPerHour = 60 * SecsPerMin
local SecsPerDay = 24 * SecsPerHour
local DaysToSave = 30

local MaxScrollItemsShown = nil --Number of Items the ScrollFrame can show at the same time
local MaxScrollItems = 50 -- Maximum number of items that can be listed

-- ScrollBar variables
local ScrollBarOffset = 0
local ScrollBarMin = 0
local ScrollBarMax = 800
local ScrollPosition = 0

local AllowRemoveAll = false

-- Local Functions
local function ItemKnownListed(item)
   local known, index = nil, nil
   item = Utils.GetItemLink(item)
   local id, name = Utils.ParseLink(item)

   if id then
      for i = 1, #GFL_ItemsNeeded do
         if id == Utils.ParseLink(GFL_ItemsNeeded[i].Item) then
            known, index = true, i
            break
         end
      end
   end

   if not known then
      for i = 1, #GFL_ItemsNeeded do
         if name == GFL_ItemsNeeded[i].Item then
            known, index = true, i
            break
         end
      end
   end

   return known, index
end

local function ItemKnownDelisted(item)
   local known, index = nil, nil
   item = Utils.GetItemLink(item)
   local id, name = Utils.ParseLink(item)

   if id then
      for i = 1, #GFL_ItemsNoLongerNeeded do
         if id == Utils.ParseLink(GFL_ItemsNoLongerNeeded[i].Item) then
            known, index = true, i
            break
         end
      end
   end

   if not known then
      for i = 1, #GFL_ItemsNoLongerNeeded do
         if name == GFL_ItemsNoLongerNeeded[i].Item then
            known, index = true, i
            break
         end
      end
   end

   return known, index 
end

local function AddonSendMessage(remove, link, date, owner)
   local messageAction
   if not remove then
      messageAction = "A=A;"
   else
      messageAction = "A=R;"
   end
   SendAddonMessage(ADDON_PREFIX, messageAction.."L="..link..";D="..date..";O="..Utils.ntos(owner)..";", "GUILD")
end

local function AddonSendMessageDirect(remove, link, date, owner, player)
   local messageAction
   if not remove then
      messageAction = "A=A;"
   else
      messageAction = "A=R;"
   end
   SendAddonMessage(ADDON_PREFIX, messageAction.."L="..link..";D="..date..";O="..Utils.ntos(owner)..";", "WHISPER", player)
end

local function ParseChat(msg)
   local action = strmatch(msg, ".*A=([^;]*).*")
   local item = strmatch(msg, ".*L=([^;]*).*")
   local date = strmatch(msg, ".*D=([^;]*).*")
   local owner = strmatch(msg, ".*O=([^;]*).*")
   date = tonumber(date)
   return action, item, date, owner
end

local function AddItem(item, date, owner)
   if #GFL_ItemsNeeded < MaxScrollItems then
      local delisted, delistedIndex = ItemKnownDelisted(item)

      if (delisted and not date) or
         (delisted and GFL_ItemsNoLongerNeeded[delistedIndex].Date <= date) or
         (not delisted) then
         local listed, listedIndex = ItemKnownListed(item)
         if not listed then
            local t  = {}
            t.Item = item
            if date then
               t.Date = date
            else
               t.Date = GetServerTime()
            end
            if owner then
               if owner ~= "NA" then
                  t.Owner = owner
               else
                  t.Owner = nil
               end
            else
               local _,class = UnitClass("player")
               local player = UnitName("player")
               local time = GetServerTime()
               t.Owner = Utils.ListToStringList({player,class,time}, ":")
            end
            table.insert(GFL_ItemsNeeded, 1, t)
            GFL_Main.ScrollFrame.OnUpdate(GFL_MainFrame_ScrollFrame)
            if not date then
               AddonSendMessage(false, t.Item, t.Date, t.Owner)
            end
         else
            if date then
               GFL_ItemsNeeded[listedIndex].Date = math.max(date, GFL_ItemsNeeded[listedIndex].Date)
            end
            if not owner then
               local ownersList = Utils.PackStringList(GFL_ItemsNeeded[listedIndex].Owner, "::")
               local _,class = UnitClass("player")
               local player = UnitName("player")
               local time = GetServerTime()
               owner = Utils.ListToStringList({player,class,time}, ":")

               local isOwner = Utils.PlayerIsOwner(ownersList)

               if not isOwner then
                  table.insert(ownersList, owner)
                  GFL_ItemsNeeded[listedIndex].Owner = Utils.ListToStringList(ownersList, "::")
                  GFL_ItemsNeeded[listedIndex].Date = GetServerTime()
                  AddonSendMessage(false, GFL_ItemsNeeded[listedIndex].Item, GFL_ItemsNeeded[listedIndex].Date, GFL_ItemsNeeded[listedIndex].Owner)
                  GFL_Main.ScrollFrame.OnUpdate(GFL_MainFrame_ScrollFrame)
               end
            else
               if owner ~= "NA"  and GFL_ItemsNeeded[listedIndex].Owner ~= owner then
                  GFL_ItemsNeeded[listedIndex].Owner = Utils.MergeStringLists(GFL_ItemsNeeded[listedIndex].Owner, owner, "::")
               end
            end
         end
      end
      if delisted then
         GFL_ItemsNoLongerNeeded[delistedIndex].Owner = Utils.RemoveOwners(GFL_ItemsNoLongerNeeded[delistedIndex].Owner, owner)
         if GFL_ItemsNoLongerNeeded[delistedIndex].Owner == "" then
            table.remove(GFL_ItemsNoLongerNeeded, delistedIndex)
         end
      end
   end
end

local function RemoveItem(item, date, owner)
   local listed, listedIndex = ItemKnownListed(item)

   if (listed and not date) or
      (listed and GFL_ItemsNeeded[listedIndex].Date < date) or
      (not listed) then
      if listedIndex then
         GFL_Main.RemoveItemByIndex(listedIndex, date, owner)
      end
   end
end

local function CanViewItem(owner)
   local canView = false
   local ownersList = Utils.PackStringList(owner, "::")
   for i=1, #ownersList do
      local owner = Utils.UnpackStringList(ownersList[i] or nil)
      if strcmputf8i(owner or "", UnitName("player")) == 0 or
         (GFL_Options.MainFrame.ShowAllChecked or false) then
            canView = true
            break
      end
   end

   if not owner and (GFL_Options.MainFrame.ShowAllChecked or false) then
      canView = true
   end

   return canView
end

local function NumItemsViewed()
   local count = 0
   for i=1, #GFL_ItemsNeeded do
      if CanViewItem(GFL_ItemsNeeded[i].Owner or nil) then
         count = count + 1
      end
   end
   return count
end

local function UpdateScrollOffset()
   local viewCount = NumItemsViewed()
   if viewCount > MaxScrollItemsShown then
      ScrollBarOffset = (viewCount - MaxScrollItemsShown) * (ScrollPosition / (ScrollBarMax - ScrollBarMin))
      if ScrollBarOffset % 1 >= .5 then
         ScrollBarOffset = math.ceil(ScrollBarOffset)
      else
         ScrollBarOffset = math.floor(ScrollBarOffset)
      end
      if not GFL_Options.MainFrame.ShowAllChecked then
         -- Need to scale offset to show items
         local count = 0
         for i=1, #GFL_ItemsNeeded do
            if CanViewItem(GFL_ItemsNeeded[i].Owner or nil) then
               count = count + 1
               if count == ScrollBarOffset then
                  ScrollBarOffset = i
                  break
               end
            end
         end
      end
   else
      ScrollBarOffset = 0
   end
end

local function ScrollFrame_Update(self)
   local name= self:GetName()
   local scrollBar = _G[name.."ScrollBar"]
   local scrollUpButton = _G[ name.."ScrollBarScrollUpButton" ];
   local scrollDownButton = _G[ name.."ScrollBarScrollDownButton" ];

   if NumItemsViewed() > MaxScrollItemsShown then
      scrollBar:Show()
   else
      scrollBar:Hide()
      scrollUpButton:Disable();
      scrollDownButton:Disable();
   end

   if scrollBar:IsShown() then
      scrollBar:SetMinMaxValues(ScrollBarMin, ScrollBarMax)

      -- Arrow button handling
      if ( scrollBar:GetValue() == ScrollBarMin) then
         scrollUpButton:Disable();
      else
         scrollUpButton:Enable();
      end
      if ((scrollBar:GetValue()) == ScrollBarMax) then
         scrollDownButton:Disable();
      else
         scrollDownButton:Enable();
      end
   end
end

-- Main Functions
GFL_Main = {}
GFL_Main.MainFrame = {}
GFL_Main.MiniMap = {}
GFL_Main.DraggingButton = {}
GFL_Main.EditBox = {}
GFL_Main.ScrollFrame = {}

GFL_ItemsNeeded = {}
GFL_ItemsNoLongerNeeded = {}
GFL_Options = {}
GFL_Options.Minimap = {}
GFL_Options.MainFrame = {}

function GFL_Main.MainFrame.OnEvent(self, event, ...)
   local args = {...}
	local currentPlayer = UnitName("player")
   

   if event == "CHAT_MSG_ADDON" and args[1] == ADDON_PREFIX then
      if args[2] then
         local action, item, date, owner = ParseChat(args[2])
         if action == "R" and strcmputf8i(args[5], currentPlayer) ~= 0 then
            --Remove
            if not owner then
               owner = "v100"
            end
            RemoveItem(item, date, owner)
         elseif action == "A" and strcmputf8i(args[5], currentPlayer) ~= 0 then
            -- Add
            AddItem(item, date, owner)
         elseif action == "S" then
            if strcmputf8i(args[5], currentPlayer) == 0 then
               for i = 1, #GFL_ItemsNeeded do
                  AddonSendMessage(false, GFL_ItemsNeeded[i].Item, GFL_ItemsNeeded[i].Date, GFL_ItemsNeeded[i].Owner)
               end

               for i = 1, #GFL_ItemsNoLongerNeeded do
                  AddonSendMessage(true, GFL_ItemsNoLongerNeeded[i].Item, GFL_ItemsNoLongerNeeded[i].Date, GFL_ItemsNoLongerNeeded[i].Owner)
               end
            else
               for i = 1, #GFL_ItemsNeeded do
                  AddonSendMessageDirect(false, GFL_ItemsNeeded[i].Item, GFL_ItemsNeeded[i].Date, GFL_ItemsNeeded[i].Owner, args[5])
               end

               for i = 1, #GFL_ItemsNoLongerNeeded do
                  AddonSendMessageDirect(true, GFL_ItemsNoLongerNeeded[i].Item, GFL_ItemsNoLongerNeeded[i].Date, GFL_ItemsNoLongerNeeded[i].Owner, args[5])
               end
            end
         elseif action == "T" then
         end
      end
   elseif event == "ADDON_LOADED" and args[1] == ADDON_NAME then
      for i = #GFL_ItemsNoLongerNeeded, 1, -1 do
         if tonumber(GFL_ItemsNoLongerNeeded[i].Date) < tonumber(GetServerTime() - (DaysToSave * SecsPerDay)) then
            table.remove(GFL_ItemsNoLongerNeeded, i)
         end
      end
      GFL_MainFrame_Checkbox:SetChecked(GFL_Options.MainFrame.ShowAllChecked)
      SendAddonMessage(ADDON_PREFIX, "A=S;", "GUILD")
      Utils.SetFramePosition(GFL_MainFrame, GFL_Options.MainFrame.xpos, GFL_Options.MainFrame.ypos, GFL_Main.MainFrame.ResetPosition)
   elseif event == "PLAYER_ENTERING_WORLD" then
      SendAddonMessage(ADDON_PREFIX, "A=S;", "GUILD")
   elseif event == "MODIFIER_STATE_CHANGED" then
      if args[1] == "LSHIFT" or args[1] == "RSHIFT" then
         if args[2] == 1 then AllowRemoveAll = true else AllowRemoveAll = false end
         GFL_Main.ScrollFrame.OnUpdate(GFL_MainFrame_ScrollFrame)
      end
   end
end

function GFL_Main.MainFrame.ResetPosition(self)
   self:SetPoint("CENTER", "UIParent")
end

function GFL_Main.RemoveItemByIndex(index, date, owner)
   if not date then
      local viewableIndex = 0
      for i=1, #GFL_ItemsNeeded do
         local itemStruc = GFL_ItemsNeeded[i]
         if CanViewItem(itemStruc.Owner or nil) then
            viewableIndex = viewableIndex + 1
            if viewableIndex == index then
               index = i
               break
            end
         end
      end
   end
         
   if index and type(index) == "number" then
      local item = GFL_ItemsNeeded[index].Item
      local t = {}
      t.Item = item
      if date then
         t.Date = date
      else
         t.Date = GetServerTime()
      end
      if owner then
         if owner == "NA" then
            t.Owner = nil
         elseif owner == "v100" then
            t.Owner = Utils.TouchOwners(GFL_ItemsNeeded[index].Owner)
         else
            t.Owner = owner
         end
      else
         if not AllowRemoveAll then
            local _, class = UnitClass("player")
            local player = UnitName("player")
            local time = GetServerTime()
            t.Owner = Utils.ListToStringList({player,class,time}, ":")
         else
            t.Owner = Utils.TouchOwners(GFL_ItemsNeeded[index].Owner)
         end
      end
      local delisted, delistedIndex = ItemKnownDelisted(item)
      if delisted then
         GFL_ItemsNoLongerNeeded[delistedIndex].Date = math.max(t.Date, GFL_ItemsNoLongerNeeded[delistedIndex].Date)
         if owner ~= "NA"  and GFL_ItemsNoLongerNeeded[delistedIndex].Owner ~= owner then
            GFL_ItemsNoLongerNeeded[delistedIndex].Owner = Utils.MergeStringLists(GFL_ItemsNoLongerNeeded[delistedIndex].Owner, owner, "::")
         end
      else
         table.insert(GFL_ItemsNoLongerNeeded, 1, t)
      end
      local cleanedList = Utils.RemoveOwners(GFL_ItemsNeeded[index].Owner, t.Owner)
      GFL_ItemsNeeded[index].Owner = cleanedList
      if GFL_ItemsNeeded[index].Owner == "" then
         table.remove(GFL_ItemsNeeded, index)
      end
      if not date then
         AddonSendMessage(true, t.Item, t.Date, t.Owner)
      end
      GFL_Main.ScrollFrame.OnUpdate(GFL_MainFrame_ScrollFrame)
   end
end

function GFL_Main.OnLoad(self)
   GFL_MainFrame_OkButton:SetFrameLevel(8)

   success = RegisterAddonMessagePrefix(ADDON_PREFIX)
   Utils.RegisterEvents(self, GFL_Main.MainFrame.OnEvent, 'CHAT_MSG_ADDON', 'ADDON_LOADED', 'PLAYER_ENTERING_WORLD', 'MODIFIER_STATE_CHANGED')
end

function GFL_Main.DraggingButton.OnLoad(self)
   self:RegisterForClicks("LeftButtonDown")
   self:RegisterForDrag("LeftButton")
end

function GFL_Main.DraggingButton.OnClick(self)
   Utils.GetCursorRelativePosition(GFL_MainFrame)
end

function GFL_Main.DraggingButton.OnDragStart(self)
   self:SetScript("OnUpdate", GFL_Main.DraggingButton.OnUpdate)
end

function GFL_Main.DraggingButton.OnDragStop(self)
   self:SetScript("OnUpdate", nil)
end

function GFL_Main.DraggingButton.OnUpdate(self)
   local x, y = Utils.GetFramePosition(GFL_MainFrame)
   Utils.SetFramePosition(GFL_MainFrame, x, y, GFL_Main.MainFrame.ResetPosition)
   GFL_Options.MainFrame.xpos = x
   GFL_Options.MainFrame.ypos = y
end

function GFL_Main.MainFrame.OnClickClose(self)
	PlaySound("igMainMenuQuit");
   Utils.ToggleFrame(GFL_MainFrame)
end

function GFL_Main.EditBox.OnClickOk(self)
   local editbox = GFL_MainFrame_EditBox
   local text = editbox:GetText()
   AddItem(text)
   editbox:SetText("")
end

function GFL_Main.EditBox.OnTextChanged(frame)
   local text = frame:GetText()
   if #text > 0 then
      GFL_MainFrame_OkButton:Show()
   else
      GFL_MainFrame_OkButton:Hide()
   end
end

function GFL_Main.EditBox.OnEnterPressed(self)
   GFL_Main.EditBox.OnClickOk(GFL_MainFrame_OkButton)
end

function GFL_Main.EditBox.OnEscapePressed(frame)
   frame:SetText("")
   frame:ClearFocus()
end

function GFL_Main.ScrollFrame.OnUpdate(self)
   if self:IsVisible() then
      ScrollFrame_Update(self)
      UpdateScrollOffset()
      local lineplusoffset=0
      local index = 0

      for line=1, MaxScrollItemsShown do
         _G["ScrollItem_"..line]:Hide()
         _G["RemoveButton_"..line]:Hide()
      end

      for line=1, #GFL_ItemsNeeded do
         lineplusoffset = line + ScrollBarOffset
         if (lineplusoffset <= #GFL_ItemsNeeded) and
            (index < MaxScrollItemsShown) then
            local itemStruc = GFL_ItemsNeeded[lineplusoffset]
            if CanViewItem(itemStruc.Owner or nil) then
               local item = Utils.GetItemLink(itemStruc.Item)
               if item ~= itemStruc.Item then
                  GFL_ItemsNeeded[lineplusoffset].Item = item
               end

               local ownersList = Utils.PackStringList(itemStruc.Owner, "::")
               local _,class = UnitClass("player")
               owner = UnitName("player")..":"..class

               local isOwner = Utils.PlayerIsOwner(ownersList)

               index = index + 1
               _G["ScrollItem_"..index]:SetText(item)
               _G["ScrollItem_"..index]:Show()
               _G["RemoveButton_"..index]:Show()
               if AllowRemoveAll or isOwner then
                  _G["RemoveButton_"..index]:Enable()
               else
                  _G["RemoveButton_"..index]:Disable()
               end
            end
         end
      end
   end
end

function GFL_Main.ScrollFrame.OnVerticalScroll(self, scrollPosition)
   ScrollPosition = scrollPosition
   UpdateScrollOffset()
   GFL_Main.ScrollFrame.OnUpdate(self)
end

function GFL_Main.OnCheckboxLoad(checkbox)
   _G[checkbox:GetName().."Text"]:SetText("Show All")
end

function GFL_Main.OnCheckboxClicked(checkbox)
   GFL_Options.MainFrame.ShowAllChecked = checkbox:GetChecked()
   GFL_Main.ScrollFrame.OnUpdate(GFL_MainFrame_ScrollFrame)
end

function GFL_Main.ScrollFrame.PopulateScrollFrame(self)
   local buffer = -10
   local index = 1
   repeat
      local i = index
      local item = self:CreateFontString("ScrollItem_"..i, "ARTWORK", "GameFontNormal")
      item:SetTextColor(1,1,1,1)
      item:SetText("Temp")
      if i == 1 then
         item:SetPoint("TOPLEFT", self, "TOPLEFT", math.abs(buffer), buffer)
      else
         item:SetPoint("TOPLEFT", _G["ScrollItem_"..(i-1)], "BOTTOMLEFT", 0, buffer)
      end
      item:Hide()

      if not MaxScrollItemsShown then
         MaxScrollItemsShown = math.floor(self:GetHeight() / (item:GetHeight() + math.abs(buffer)))
      end

      local removeButton = CreateFrame("Button", "RemoveButton_"..i, self, "UIPanelButtonTemplate")
      removeButton:SetText("X")
      removeButton:SetPoint("TOP", item, "TOP", 0, 5)
      removeButton:SetPoint("RIGHT", self, "RIGHT", -10, 0)
      removeButton:SetScript("OnClick", function () GFL_Main.RemoveItemByIndex(i) end)
      removeButton:Hide()

      index = index + 1
   until index == (MaxScrollItemsShown + 1)
   self:SetPropagateKeyboardInput(true)
end

function GFL_Main.MiniMap.OnEvent(self, event, ...)
   local args = {...}
   if event == "ADDON_LOADED" and args[1] == ADDON_NAME then
      Utils.SetFramePosition(GFL_MMButton, GFL_Options.Minimap.xpos, GFL_Options.Minimap.ypos, GFL_Main.MiniMap.ResetPosition)
   end
end

function GFL_Main.MiniMap.OnLoad(self)
   self:RegisterForClicks("LeftButtonDown", "LeftButtonUp")
   self:RegisterForDrag("LeftButton")
   self:SetFrameLevel(8)
   Utils.RegisterEvents(self, GFL_Main.MiniMap.OnEvent, 'ADDON_LOADED')
end

function GFL_Main.MiniMap.OnClick(self, button, down)
   if down then
      GFL_UtilFuncs.GetCursorRelativePosition(self)
   else
      GFL_UtilFuncs.ToggleFrame(GFL_MainFrame);
   end
end

function GFL_Main.MiniMap.OnEnter(self)
   GFL_UtilFuncs.ShowToolTip(self)
end

function GFL_Main.MiniMap.OnLeave(self)
end

function GFL_Main.MiniMap.OnDragStart(self, button)
   self:LockHighlight()
   Utils.TooltipRelease()
   self:SetScript("OnEnter", nil)
   self:SetScript("OnUpdate", GFL_Main.MiniMap.OnUpdate)
end

function GFL_Main.MiniMap.OnDragStop(self)
   self:UnlockHighlight()
   self:SetScript("OnUpdate", nil)
   self:SetScript("OnEnter", GFL_UtilFuncs.ShowToolTip)
end

function GFL_Main.MiniMap.OnUpdate(self)
   local x, y = GFL_UtilFuncs.GetFramePosition(self)
   GFL_UtilFuncs.SetFramePosition(self, x, y, GFL_Main.MiniMap.ResetPosition)
   GFL_Options.Minimap.xpos = x
   GFL_Options.Minimap.ypos = y
end

function GFL_Main.MiniMap.ResetPosition(self)
   self:SetPoint("CENTER", "Minimap", "CENTER", -64.9, -37.5) 
end
