GFL_UtilFuncs = {}
GFL_ItemDictionary = {}

local ClassColors = {}
ClassColors["DEATHKNIGHT"] = 0xC41F3B
ClassColors["DEMONHUNTER"] = 0xA330C9
ClassColors["DRUID"]       = 0xFF7D0A
ClassColors["HUNTER"]      = 0xABD473
ClassColors["MAGE"]        = 0x69CCF0
ClassColors["MONK"]        = 0x00FF96
ClassColors["PALADIN"]     = 0xF58CBA
ClassColors["PRIEST"]      = 0xFFFFFF
ClassColors["ROGUE"]       = 0xFFF569
ClassColors["SHAMAN"]      = 0x0070DE
ClassColors["WARLOCK"]     = 0x9482C9
ClassColors["WARRIOR"]     = 0xC79C6E

local Tooltip, RequesterTooltip
local LibQTip = LibStub('LibQTip-1.0')

local CursorRelativeX = 0 -- Horizontal distance from cursor to TOPLEFT of MainFrame
local CursorRelativeY = 0 -- Vertical distance from cursor to TOPLEFT of MainFrame

function GFL_UtilFuncs.GetClassRGB(class)
   if not class then
      return
   end
   local colors = {}
   local color = ClassColors[class]
   
   if not color then
      return
   end

   for i=1,3 do
      table.insert(colors, 1, bit.band(color, 0x0000FF)/0xFF)
      color = bit.rshift(color, 8)
   end
   return unpack(colors)
end

function GFL_UtilFuncs.PackStringList(stringList, delimiter)
   if not stringList then
      return {}
   end

   if not delimiter then
      delimiter = ":"
   end

   local start = 1
   local list = {}
   
   while true do
      local first, last = strfind(stringList, delimiter, start)
      if not first then
         table.insert(list, strsub(stringList, start))
         break
      end
      table.insert(list, strsub(stringList, start, first - 1))
      start = last + 1
   end
   return list
end

function GFL_UtilFuncs.UnpackStringList(stringList, delimiter)
   return unpack(GFL_UtilFuncs.PackStringList(stringList, delimiter))
end

function GFL_UtilFuncs.ListToStringList(list, delimiter)
   if not list then
      return
   end

   if not delimiter then
      delimiter = ":"
   end

   local string = ""
   for i=1, #list do
      if i ~= #list then
         string = string..list[i]..delimiter
      else
         string = string..list[i]
      end
   end
   return string
end

function GFL_UtilFuncs.PlayerIsOwner(ownerList, player)
   if not player then
      player = UnitName("player")
   end

   for i=1, #ownerList do
      local o = GFL_UtilFuncs.PackStringList(ownerList[i], ":")
      if strcmputf8i(player, o[1]) == 0 then
         return true, i
      end
   end
   return false, nil
end

function GFL_UtilFuncs.TableContains(table, string)
   for _, v in ipairs(table) do
      if strcmputf8i(v, string) == 0 then
         return true
      end
   end
   return false
end

function GFL_UtilFuncs.TouchOwners(ownerList)
   local pol = GFL_UtilFuncs.PackStringList(ownerList, "::")
   local time = GetServerTime()
   for i=1, #pol do
      local o = GFL_UtilFuncs.PackStringList(pol[i], ":")
      pol[i] = GFL_UtilFuncs.ListToStringList({o[1],o[2],time}, ":")
   end
   return GFL_UtilFuncs.ListToStringList(pol, "::")
end

function GFL_UtilFuncs.RemoveOwners(ownerList, ownersToRemove)
   if not ownersToRemove then
      return ownerList
   end

   local potr = GFL_UtilFuncs.PackStringList(ownersToRemove, "::")
   local pol = GFL_UtilFuncs.PackStringList(ownerList, "::")
   for _,v in ipairs(potr) do
      local player,_,date = GFL_UtilFuncs.UnpackStringList(v, ":")
      local isOwner, index = GFL_UtilFuncs.PlayerIsOwner(pol, player)
      if isOwner then
         local _,_,pol_date = GFL_UtilFuncs.UnpackStringList(pol[index], ":")
         if date > pol_date then
            table.remove(pol, index)
         end
      end
   end
   return GFL_UtilFuncs.ListToStringList(pol, "::")
end

function GFL_UtilFuncs.MergeStringLists(sl1, sl2, delimiter)
   if not sl1 then
      return sl2
   end

   if not sl2 then
      return sl1
   end

   if not delimiter then
      delimiter = ":"
   end

   StringList1 = GFL_UtilFuncs.PackStringList(sl1, delimiter)
   StringList2 = GFL_UtilFuncs.PackStringList(sl2, delimiter)

   for _, v in ipairs(StringList2) do
      local player,_,date = GFL_UtilFuncs.UnpackStringList(v, ":")
      local isOwner, index = GFL_UtilFuncs.PlayerIsOwner(StringList1, player)
      if not isOwner then
         table.insert(StringList1, v)
      else
         local existing = GFL_UtilFuncs.PackStringList(StringList1[index], ":")
         if date > existing[3] then
            StringList1[index] = GFL_UtilFuncs.ListToStringList({existing[1],existing[2],date},":")
         end
      end
   end

   return GFL_UtilFuncs.ListToStringList(StringList1, delimiter)
end

function GFL_UtilFuncs.ntos(o)
   if not o then
      return "NA"
   end
   return o
end

function GFL_UtilFuncs.RegisterEvents(self, callback, ...)
   for i = 1, select('#', ...) do
      self:RegisterEvent(select(i, ...))
   end
   self:SetScript("OnEvent", callback)
end

function GFL_UtilFuncs.ParseLink(link)
   if not link then
      return
   end

   local id = strmatch(link, ".*Hitem:([^:]*)")
   local name = strmatch(link, ".*h%[([^%]]*)")

   if not id then
      return id, link
   end

   return id, name
end

function GFL_UtilFuncs.GetItemLink(item)
   local _, link = GetItemInfo(item)
   local id, name = GFL_UtilFuncs.ParseLink(link)

   if not id and GFL_ItemDictionary[name] then
      return GFL_ItemDictionary[name]
   end

   if name and not GFL_ItemDictionary[name] then
      GFL_ItemDictionary[name] = link
   end

   return link or GFL_ItemDictionary[item] or item
end

function GFL_UtilFuncs.SetFramePosition(frame, x, y, resetFunc)
   if not frame then
      return
   end

   if not x or not y then
      if resetFunc and type(resetFunc) == "function" then
         resetFunc(frame)
      end
   else
      frame:SetPoint("TOPLEFT", "UIParent", "BOTTOMLEFT", x, y)
   end
end

function GFL_UtilFuncs.GetFramePosition(frame)
   if not frame then
      return
   end

   local xpos, ypos = GetCursorPosition()
   local scale = frame:GetEffectiveScale()

   local x = (xpos / scale) - CursorRelativeX
   local y = (ypos / scale) - CursorRelativeY

   return x, y
end

function GFL_UtilFuncs.ShowRequesterTooltip(self, arg)
   if RequesterTooltip then LibQTip:Release(RequesterTooltip); RequesterTooltip = nil end
   local tt = LibQTip:Acquire("GFL_RequesterTooltip", 1, "LEFT")
   tt:Clear()
   RequesterTooltip = tt
   tt:AddLine('Requesters:')
   tt:SetLineTextColor(1, 1, .82, 0, 1)
   tt:SetAutoHideDelay(0.1, self, GFL_UtilFuncs.RequesterTooltipRelease)
   tt:SetBackdropColor(0, 0, 0, 0.5)
   tt:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)
   tt:SetPoint("TOP", self, "TOP")
   local xpos = GetCursorPosition()
   xpos = xpos * UIParent:GetEffectiveScale()
   if xpos <= 400 then
      tt:SetPoint("LEFT", Tooltip, "RIGHT")
   else
      tt:SetPoint("RIGHT", Tooltip, "LEFT")
   end
   tt:SetClampedToScreen(true)

   for i=1, #arg do
      local owner = GFL_UtilFuncs.PackStringList(arg[i])
      local r, g, b = GFL_UtilFuncs.GetClassRGB(owner[2] or nil)
      tt:AddLine(owner[1])
      tt:SetLineTextColor(i + 1, r, g, b, 1)
   end
   tt:Show()
end

function GFL_UtilFuncs.RequesterTooltipRelease()
   if RequesterTooltip then
      LibQTip:Release(RequesterTooltip)
      RequesterTooltip = nil
   end
end

function GFL_UtilFuncs.ShowToolTip(frame)
   if not frame then
      return
   end

   if Tooltip then LibQTip:Release(Tooltip); Tooltip = nil end
   local tt = LibQTip:Acquire("GFL_Tooltip", 2, "LEFT", "RIGHT")
   Tooltip = tt
   tt:AddLine('Guild Currently Needs:')
   tt:SetAutoHideDelay(0.1, frame, GFL_UtilFuncs.TooltipRelease)
   tt:SetBackdropColor(0, 0, 0, 0.5)
   tt:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)

   for i=1, #GFL_ItemsNeeded do
      item = GFL_UtilFuncs.GetItemLink(GFL_ItemsNeeded[i].Item)
      if item ~= GFL_ItemsNeeded[i].Item then
         GFL_ItemsNeeded[i].Item = item
      end
      local ownersList = GFL_UtilFuncs.PackStringList(GFL_ItemsNeeded[i].Owner, "::")
      local owner = GFL_UtilFuncs.PackStringList(ownersList[1] or nil)
      local r,g,b = GFL_UtilFuncs.GetClassRGB(owner[2] or nil)
      
      tt:AddLine()
      tt:SetCell(i + 1, 1, item)
      local ownerStr = owner[1]
      if #ownersList > 1 then
         ownerStr = ownerStr.."(+)"
         tt:SetLineScript(i + 1, "OnEnter", GFL_UtilFuncs.ShowRequesterTooltip, ownersList)
      end
      tt:SetCell(i + 1, 2, ownerStr)
      tt:SetCellTextColor(i + 1, 2, r, g, b, 1)
   end
   tt:AddLine(' ')
   tt:AddSeparator(1, 1, .82, 0, 1)
   tt:AddLine('(+) Hover to see Requester List')
   tt:SetColumnTextColor(1, 1, .82, 0, 1)
   tt:SetPoint("TOPRIGHT", frame, "BOTTOMRIGHT")
   --tt:UpdateScrolling(400)
   tt:SetClampedToScreen(true)
   tt:Show()
end

function GFL_UtilFuncs.TooltipRelease()
   if Tooltip then
      LibQTip:Release(Tooltip)
      Tooltip = nil
   end
end

function GFL_UtilFuncs.ToggleFrame(frame)
   if not frame then
      return
   end

   PlaySound("igMainMenuQuit");
   if(frame:IsShown()) then
      frame:Hide()
   else
      frame:Show()
   end
end

function GFL_UtilFuncs.GetCursorRelativePosition(frame)
   if not frame then
      return
   end

   local xpos, ypos = GetCursorPosition()
   local xOffset = frame:GetLeft()
   local yOffset = frame:GetTop()
   local scale = frame:GetEffectiveScale()

   CursorRelativeX = (xpos / scale) - xOffset 
   CursorRelativeY = (ypos / scale) - yOffset
end
